#include "stdafx.h"
#include "xmlhelper.h"

bool a2i(const char* str, int *result)
{
    assert(str && result);
    if (strspn(str, "-0123456789") != strlen(str)) return false;
    *result = atoi(str);
    return true;
}

void i2a(int value, std::string* res)
{
    assert(res);
    char sup_conversion[16];
    sprintf(sup_conversion, "%d", value);
    res->assign(sup_conversion);
}

xml::node::node() : m_Node(NULL) {}
xml::node::node(xnode xml_node) : m_Node(xml_node) {}
xml::node::~node() {}

bool xml::node::load(xml::node parent_node, const char *filename)
{
    assert(filename);
    bool res = false;

    if (parent_node)    
        m_Node = xmlLoadToNode(parent_node, filename);
    else
        m_Node = xmlLoad(filename);
    if (m_Node) res = true;
    return res;
}

bool xml::node::createnew(xml::node parent_node, const char *node_name)
{
    assert(node_name);
    bool res = false;
    if (parent_node)
        m_Node = xmlCreateNode(parent_node, node_name);
    else
        m_Node = xmlCreateRootNode(node_name);
    if (m_Node) res = true;
    return res;
}

bool xml::node::save(const char *filename)
{
    assert(filename);
    bool res = false;
    if (m_Node)
    {
        if (xmlSave(m_Node, filename))
            res = true;
    }

    return res;
}

void xml::node::getname(std::string* value) const
{
    assert(value);
    xstring val = xmlGetName(m_Node);
    if (val) *value = val;
    else value->clear();
}

bool xml::node::getattribute(const char* attribute, std::string* value) const
{
    assert(m_Node > 0 && attribute && value);

    bool res = false;
    xstring val = xmlGetAttribute(m_Node, attribute);
    if (val) { *value = val; res = true; }
    else value->clear();
    return res;
}

bool xml::node::getattribute(const char* attribute, int* value) const
{
    assert(m_Node > 0 && attribute && value);

    bool res = false;
    xstring str = xmlGetAttribute(m_Node, attribute);
    int tmp = 0;
    if (str && a2i(str, &tmp)) { *value = tmp; res = true; }
    return res;
}

void xml::node::gettext(std::string *text) const
{
    assert(m_Node > 0 && text);

    xstring txt = xmlGetText(m_Node);
    if (txt) *text = txt;
    else text->clear();
}

void xml::node::getallsubnodestext(const char* subnode, std::string* text) const
{
    assert(m_Node > 0 && subnode && text);
    
    text->clear();
    xlist lst = xmlRequest(m_Node, subnode);    
    for (int i=0, sz = xmlGetListSize(lst); i < sz; ++i)
    {
        xstring txt = xmlGetText(lst[i]);
        if (txt)
            text->append(txt);
    }

    xmlFreeList(lst);
}

void xml::node::getsubnodetext(const char* subnode, std::string* text) const
{
    assert(m_Node > 0 && subnode && text);

    text->clear();
    xlist lst = xmlRequest(m_Node, subnode);
    if (xmlGetListSize(lst) > 0)
    {
        xstring txt = xmlGetText(lst[0]);
        if (txt)
            text->assign(txt);
    }

    xmlFreeList(lst);
}

void xml::node::getcdata(std::string *text) const
{
    assert (m_Node > 0 && text);
    xstring txt = xmlGetCDataText(m_Node);
    if (txt)
        text->assign(txt);
    else
        text->clear();
}

void xml::node::deletesubnodes()
{
    if (m_Node)
    {
        xml::request ns(m_Node, "*");
        ns.deletenodes();
    }
}

void xml::node::deletenode()
{
    if (m_Node)
      xmlDeleteNode(m_Node);
    m_Node = 0;
}

bool xml::node::setattribute(const char* attribute, const std::string &value) const
{
    assert (m_Node && attribute);
    return (xmlSetAttribute(m_Node, attribute, value.c_str()) != 0);
}

bool xml::node::setattribute(const char* attribute, int value) const
{
    assert (m_Node && attribute);
    std::string num;
    i2a(value, &num);
    return (xmlSetAttribute(m_Node, attribute, num.c_str()) != 0);
}

bool xml::node::settext(const std::string &value) const
{
    assert(m_Node);
    return (xmlSetText(m_Node, value.c_str()) != 0);
}

bool xml::node::setsubnodetext(const char* subnode, const std::string& text) const
{
    assert(m_Node && subnode);

    bool res = false;
    xlist lst = xmlRequest(m_Node, subnode);
    if (xmlGetListSize(lst) > 0)
        res = (xmlSetText(lst[0], text.c_str()) != 0);    
    else
    {
        xnode sn = xmlCreateNode(m_Node, subnode);
        if (sn)
            res = (xmlSetText(sn, text.c_str()) != 0);
    }

    return res;
}

xml::node xml::node::createsubnode(const char* name) const
{
    assert(m_Node && name);
    return xmlCreateNode(m_Node, name);
}

xml::node xml::node::makecopytochild(xml::node parent_node) const
{
    assert(m_Node && parent_node);
    return xmlMakeNodeCopy(parent_node, m_Node);
}

xml::request::request(const char *request_string) : m_NodeList(NULL), m_ListSize(0)
{
    assert(request_string);
    m_NodeList = xmlRootRequest(request_string);
    if (m_NodeList)
    {
        m_ListSize = xmlGetListSize(m_NodeList);
        if (!m_ListSize) { xmlFreeList(m_NodeList); m_NodeList = NULL; }
    }
}

xml::request::request(xml::node request_node, const char *request_string) : m_NodeList(NULL), m_ListSize(0)
{
    assert(request_string);
    m_NodeList = xmlRequest(request_node, request_string);
    if (m_NodeList)
    {
        m_ListSize = xmlGetListSize(m_NodeList);
        if (!m_ListSize) { xmlFreeList(m_NodeList); m_NodeList = NULL; }
    }
}

xml::request::~request()
{
    if (m_NodeList) 
        xmlFreeList(m_NodeList);
}

const xml::node xml::request::operator[](int node_index)
{
    return (node_index >= 0 && node_index < m_ListSize) ? m_NodeList[node_index] : 0;    
}

void xml::request::deletenodes()
{
    for (int i=0, e=m_ListSize; i<e; ++i)
        xmlDeleteNode(m_NodeList[i]);
    if (m_NodeList) 
        xmlFreeList(m_NodeList);
    m_NodeList = NULL;
    m_ListSize = 0;
}
