
#include "wrapper.h"
#include <libxml/xmlsave.h>
#include <libxml/xpath.h>
#include <encoding.h>
#include <vector>
//------------------------------------------------------------------------------------
CXMLWrapper::CXMLWrapper() : rootdoc(NULL), rootnode(NULL)
{
	//  
	xmlInitParser();

	//       - Root,
	//      xml ,   xml-	
	rootdoc = xmlNewDoc((xmlChar*)"1.0");
	rootnode = xmlNewDocNode(rootdoc, NULL, (xmlChar*)"root", NULL);
    xmlDocSetRootElement(rootdoc, rootnode);
}

CXMLWrapper::~CXMLWrapper()
{
	//       
	Destroy();
	xmlFreeDoc(rootdoc);
	xmlCleanupParser();
}

xnode CXMLWrapper::LoadXML(const char *buffer, int size)
{
    if (!buffer || size <= 0) 
        return NULL;

    xmlDocPtr doc = xmlReadMemory(buffer, size, NULL, "UTF-8", XML_PARSE_NOERROR | XML_PARSE_NOWARNING);
    if (!doc) return NULL;

    xmlNodePtr newnode = xmlCopyNode(xmlDocGetRootElement(doc), 1);
    xmlFreeDoc(doc);
    FormatTree(newnode);

    if (!xmlAddChild(rootnode, newnode))
    { xmlFreeNode(newnode); newnode = NULL; }

    return newnode;
}

xnode CXMLWrapper::LoadXMLToNode(xnode node, const char *buffer, int size)
{
    if (!buffer || size <= 0 || !node) 
        return NULL;

    xmlDocPtr doc = xmlReadMemory(buffer, size, NULL, "UTF-8", XML_PARSE_NOERROR | XML_PARSE_NOWARNING);
    if (!doc) return NULL;

    xmlNodePtr newnode =  xmlCopyNode(xmlDocGetRootElement(doc), 1);  
    xmlFreeDoc(doc);
    FormatTree(newnode);

    if (!xmlAddChild(node, newnode))
    { xmlFreeNode(newnode); newnode = NULL; }

    return newnode;
}

xnode CXMLWrapper::LoadXMLFile(const char *filename)
{
	if (!filename) return NULL;
	
    xmlDocPtr doc = xmlReadFile(filename, "UTF-8", XML_PARSE_NOERROR | XML_PARSE_NOWARNING);
    if (!doc) return NULL;
            
    xmlNodePtr newnode = xmlCopyNode(xmlDocGetRootElement(doc), 1);
    xmlFreeDoc(doc);
    FormatTree(newnode);
    
    if (!xmlAddChild(rootnode, newnode))
        { xmlFreeNode(newnode); newnode = NULL; }
	
	return newnode;
}

xnode CXMLWrapper::LoadXMLFileToNode(xnode node, const char *filename)
{
	if (!filename || !node) return NULL;
	
    xmlDocPtr doc = xmlReadFile(filename, "UTF-8", XML_PARSE_NOERROR | XML_PARSE_NOWARNING);   
	if (!doc) return NULL;

    xmlNodePtr newnode =  xmlCopyNode(xmlDocGetRootElement(doc), 1);  
    xmlFreeDoc(doc);
    FormatTree(newnode);

    if (!xmlAddChild(node, newnode))
	       { xmlFreeNode(newnode); newnode = NULL; }
		
	return newnode;
}

int CXMLWrapper::SaveXMLFile(xnode node, const char *filename)
{
   if (!filename || !node) return 0;
       
   int res = 0;
   xmlBufferPtr buff = xmlBufferCreate();
   xmlOutputBufferPtr output_buffer = xmlOutputBufferCreateBuffer(buff, NULL);
   xmlNodePtr oldrootnode = xmlDocSetRootElement(rootdoc, node);   
   int output_stream_len = xmlSaveFormatFileTo(output_buffer, rootdoc, "UTF-8", 1);
   xmlDocSetRootElement(rootdoc, oldrootnode);
   if (output_stream_len != -1)
   {
      FILE* file = fopen(filename, "wb");
      if (file != NULL)
      {
         fwrite(buff->content, 1, output_stream_len, file);
         fflush(file);
         fclose(file);
         res = 1;
      }
   }

   if (buff)
      xmlBufferFree(buff);

   return res;
}

void CXMLWrapper::FormatTree(xmlNodePtr root)
{
    char symbols[4] = { 0x20, 0xd, 0xa, 0 };
    std::vector<xmlNodePtr> vNodesStack;
    while (root)
    {
        if (root->children)
             vNodesStack.push_back(root->children);
        
        xmlNodePtr tofree = NULL;
        if (root->type == XML_TEXT_NODE && root->content)
        {
           //   
           //   ,        
           char *ptr = (char*)root->content;
           if (strspn(ptr, symbols) == strlen(ptr))
           {               
               tofree = root;               
           }
        }
        
        root = root->next;

        if (tofree)
        {
            xmlUnlinkNode(tofree);
            xmlFreeNode(tofree);
        }

        if (!root)
        {
            int last = vNodesStack.size() - 1;
            if (last >= 0) { root = vNodesStack[last]; vNodesStack.pop_back(); }
        }
    }
}

void CXMLWrapper::Destroy()
{
	xmlNodePtr temp = rootnode->children;
	while (temp != NULL)	
	{
		xmlNodePtr ptr = temp;
        temp = temp->next;
		xmlUnlinkNode(ptr);
 		xmlFreeNode(ptr);
	}
 
}

xlist CXMLWrapper::Request(xnode node, const char *request)
{
	if (!request) return NULL;

	xmlNodePtr reqnode = rootnode;
	if (node)
		reqnode = node;
	if (!reqnode) return NULL;

	xmlXPathContextPtr xpathCtx = xmlXPathNewContext(rootdoc);
	if (!xpathCtx) return NULL;
	xpathCtx->node = reqnode;

	xmlChar *xcRequest = xmlCharStrdup(request);
	xmlXPathObjectPtr xpathObj = xmlXPathEval(xcRequest, xpathCtx);

	if (!xpathObj)
	{
		xmlXPathFreeContext(xpathCtx);
		return NULL;
	}

	if(xpathObj->type != XPATH_NODESET)
	{
		xmlXPathFreeObject(xpathObj);
		xmlXPathFreeContext(xpathCtx);
		return NULL;
	}

	xlist newlist = NULL;
	xmlNodeSetPtr nodeset = xpathObj->nodesetval;
	if (nodeset && nodeset->nodeNr > 0)
	{
		int nodes = nodeset->nodeNr;
		newlist = new xnode[nodes + 1];
		if (newlist)
		{
			for (int i=0; i<nodes; i++)			
                newlist[i] = xpathObj->nodesetval->nodeTab[i];			
			newlist[nodes] = 0; //   
		}
	}

	xmlXPathFreeObject(xpathObj);
	xmlXPathFreeContext(xpathCtx);
	free(xcRequest);
	
	return newlist;
}

int CXMLWrapper::GetListSize(xlist nodelist)
{
	if (!nodelist) return 0;
	int sz = 0;
	while (nodelist[sz] != 0) sz++;
	return sz;
}

void CXMLWrapper::ReleaseList(xlist nodelist)
{
	if (!nodelist) return;
    delete []nodelist;
}

xstring CXMLWrapper::GetName(xnode node)
{
    return (node) ? node->name : NULL;	
}

xstring CXMLWrapper::GetAttribute(xnode node, const char *atname)
{
	if (!atname || !node) return NULL;
	
	xmlChar *attribute = xmlCharStrdup(atname);
	xmlAttr *tmp = NULL;
	for (xmlAttr* attr = node->properties; attr; attr = attr->next)
	{
		if (!xmlStrcmp(attribute, attr->name))
		{
			tmp = attr; break;
		}
	}

	free(attribute);
	return (tmp) ? tmp->children->content : NULL;
}

xstring CXMLWrapper::GetText(xnode node)
{	
	if (!node || !node->children) return NULL;
	return node->children->content;
}

xstring CXMLWrapper::GetCDataText(xnode node)
{    
    if (!node) return NULL;
    node = node->children; 
    while (node && node->type !=  XML_CDATA_SECTION_NODE)    
        node=node->next;
    
    return (node) ? node->content : NULL;    
}


xnode CXMLWrapper::CreateNode(xnode parentnode, const char* nodename)
{
	if (!nodename) return NULL;

	xmlNodePtr ptr = rootnode;
	if (parentnode)
		ptr = parentnode;
	if (!ptr) return NULL;
		
	xmlChar *nname = xmlCharStrdup(nodename);
	xmlNodePtr nptr = xmlNewChild(ptr, NULL, nname, NULL);
	
	free(nname);
	return nptr;
}

void CXMLWrapper::DeleteNode(xnode node)
{    	
	if (!node) return;	
	xmlUnlinkNode(node);
	xmlFreeNode(node);	
}

int CXMLWrapper::SetAttribute(xnode node, const char *atname, const char *value)
{
	if (!atname || !value || !node) return 0;
	
	xmlChar *an = xmlCharStrdup(atname);
	xmlChar *v = xmlCharStrdup(value);
	
	xmlAttr *new_attr = xmlNewProp(node, an, v);
	
	free(an);
	free(v);

	return (new_attr) ? 1 : 0;
}

int CXMLWrapper::SetText(xnode node, const char *text)
{
	if (!node) return 0;
	
	xmlChar *an = xmlCharStrdup((text) ? text : "");
	xmlNodeSetContent(node, an);	
	free(an);
	return 1;
}

xnode CXMLWrapper::MakeNodeCopy(xnode node, xnode parent_node)
{
    xmlNodePtr new_node = xmlCopyNode(node, 1);
    if (new_node)
    {        
        if (!xmlAddChild(parent_node, new_node)) 
        {
            xmlUnlinkNode(new_node);
            xmlFreeNode(new_node);
            new_node = NULL;
        }     
    }

    return new_node;
}
