#include "stdafx.h"
#include "inputProcessor.h"

InputCommand::InputCommand(tstring cmd)
{
    // trim command
    tstring_trim(&cmd);
    if (cmd.empty())
        return;

    // save trimmed command
    full_command.assign(cmd);
    
    // divide cmd for cmd+parameters
    int pos = cmd.find(L' ');
    if (pos == -1)
    {
        command.assign(cmd);
        return;
    }
    command.assign(cmd.substr(0,pos));
    parameters.append(cmd.substr(pos+1));
    tstring_trimleft(&parameters);

    // divide parameters
    const WCHAR* p = cmd.c_str() + pos + 1;
    const WCHAR* e = wcschr(p, L' ');
    while (e != NULL)
    {
        if (p != e)
            parameters_list.push_back( tstring(p, e-p) );
        p = e + 1;
        e = wcschr(p, L' ');
    }
    parameters_list.push_back(p);
}

InputProcessor::InputProcessor(PropertiesData *pdata) : pData(pdata)
{
}

InputProcessor::~InputProcessor()
{
    struct{ void operator() (InputCommand* cmd) { delete cmd; }} autodel;
    std::for_each(commands.begin(), commands.end(), autodel);
}

void InputProcessor::process(const tstring& cmd)
{    
    // process separators
    processSeparators(cmd, &commands);   

    // process aliases
    int queue_size = commands.size();
    if (queue_size == 0) 
        return;
    
    // to protect from loops in aliases
    std::vector<tstring> loops_hunter;
    for (int i=0; i<queue_size;)    
    {
        loops_hunter.push_back(commands[i]->command);
        tstring alias_cmd;
        if (processAliases(commands[i], &alias_cmd))
        {
            InputCommandsList new_cmd_list;
            processSeparators(alias_cmd, &new_cmd_list);

            bool loop = false;
            for (int j=0,je=new_cmd_list.size(); j<je; ++j)
            {
                if (std::find(loops_hunter.begin(), loops_hunter.end(), new_cmd_list[j]->command) !=
                    loops_hunter.end())
                        { loop = true; break; }
            }
            
            InputCommand *icmd = commands[i];
            commands.erase(commands.begin() + i);
            delete icmd;
            commands.insert(commands.begin() + i, new_cmd_list.begin(), new_cmd_list.end());
            queue_size = commands.size();

            if (loop)
            {
                //todo Loop in aliases !!!
                i++;
                loops_hunter.clear();
            }
        }
        else
        {
            i++;
            loops_hunter.clear();
        }
    }
}

void InputProcessor::processSeparators(const tstring& sep_cmd, InputCommandsList* result)
{
    // truncate to separate commands
    WCHAR separator = pData->cmd_separator;
    size_t pos0 = 0;
    size_t pos = sep_cmd.find(separator);
    while (pos != tstring::npos)
    {
        InputCommand *icmd = new InputCommand( sep_cmd.substr(pos0, pos-pos0) );
        result->push_back(icmd);        
        pos0 = pos+1;
        pos = sep_cmd.find(separator, pos0);
    }
    InputCommand *icmd = new InputCommand( sep_cmd.substr(pos0) );
    result->push_back(icmd);
}

bool InputProcessor::processAliases(InputCommand* cmd, tstring* result)
{
    // skip empty commands
    if (cmd->command.empty())
        return false;

    // skip system commands
    WCHAR prefix = pData->cmd_prefix;
    if (cmd->command.at(0) == prefix)
        return false;

    int aliases = pData->aliases.size();
    bool alias_found = false;
    for (int i=0; i<aliases; ++i)
    {
        if (cmd->command == pData->aliases.get(i).key)
        {
            const tstring& v = pData->aliases.get(i).value;            
            processParameters(v, cmd, result);
            alias_found = true;
            if (v == cmd->command)             
                alias_found = false;
            break;            
        }
    }    
    return alias_found;
}

void InputProcessor::processParameters(const tstring& cmd, InputCommand* params, tstring* result)
{
    // find parameters
    ParamsHelper ph(cmd);
    if (!ph.getSize())
    {
        result->assign(cmd);
        result->append(params->parameters);
        return;
    }

    result->clear();
    int params_count = params->parameters_list.size();
   
    int ptr = 0;
    int size = ph.getSize();
    for (int i=0; i<size; ++i)
    {
        int begin = ph.getFirst(i);
        result->append(cmd.substr(ptr, begin-ptr));
        int end = ph.getLast(i);

        int parameter_index = ph.getId(i);
        if (parameter_index < params_count)
        {
            const tstring& p = (parameter_index == 0) ? params->parameters : 
                params->parameters_list[parameter_index-1];
            result->append(p);
        }
        else
        {
            if (begin > 0 && cmd.at(begin-1) == L' ' && cmd.at(end) == L' ')
                end = end + 1;
        }
        ptr = end;
    }
    result->append(cmd.substr(ptr));
}
